#ifndef qsieve_header
#define qsieve_header

typedef list<int> TIntSet_Basis; 
class TIntSet : public TIntSet_Basis
// set<int,less<int> > 
// list<int>
{
 public:

  bool ist_sortiert()
    {
      iterator pos = begin();
      while (pos != end())
	{
	  iterator prev=pos; ++pos;
	  if (pos==end()) return true;
	  if (*prev >= *pos) return false;
	}
      return true;
    }
  
  void check_set(char* s) // fr Debuggingzwecke
    {
      if (ist_sortiert()) return;
      cerr << s << endl;
      for(TIntSet::iterator p=begin(); p!=end(); ++p) cerr << *p << " ";
      cerr << endl;
      exit(1);
    }
  
  inline void insert(int i)
    {       
      //check_set("Fehler vor einstelligem Insert");
      TIntSet::iterator h = end();
      if ( !empty() && *--h>=i )
	{
	  cerr << "Eigentlich nur aufsteigendes Einfuegen effizient!" << endl;
	  TIntSet::iterator pos = begin();
	  int z=0;
	  while (*pos<i) { ++pos; z++; }
	  cout << "benoetigte Schritte: " << z << endl;
	  insert(pos,i);
	}
      else push_back(i);
      //check_set("Fehler nach einstelligem Insert");
    }
  
  inline void insert(TIntSet::iterator pos, int i)
    {      
      //check_set("Fehler vor zweistelligem Insert");
      TIntSet_Basis::insert(pos,i);
      //check_set("Fehler nach zweistelligem Insert");
    }
  
  
  inline void insert(TIntSet::iterator pos,
		     TIntSet::const_iterator firstpos,
		     TIntSet::const_iterator lastpos)
    {      
      //check_set("Fehler vor dreistelligem Insert");
      TIntSet_Basis::insert(pos,firstpos,lastpos);
      //check_set("Fehler nach dreistelligem Insert");
    }
  

  inline void kombinieren(const TIntSet &IntSet2)
    /* symmetrische Differenz aus dieser Menge und der bergebenen
       Menge in dieser Menge ablegen */
    {
      TIntSet::iterator p1 = begin();
      TIntSet::const_iterator p2 = IntSet2.begin();
      TIntSet::const_iterator h2;
      TIntSet::iterator h;
      while (p1!=end())
	{
	  if ( p2!=IntSet2.end() && *p2<*p1 )
	    {
	      h2=p2;
	      do ++p2; while ( p2!=IntSet2.end() && *p2<*p1 );
	      insert(p1,h2,p2);
	    }
	  if ( p2!=IntSet2.end() && *p2==*p1 ) { h=p1++; erase(h); ++p2; }
	  else ++p1;
	}
      if (p2!=IntSet2.end()) insert(p1,p2,IntSet2.end());
      //check_set("Fehler beim kombinieren!");
    }  
};


const int no_factor = -1;

class CGleichungszeile
{
 public:
  int relevanter_Faktor;
  TIntSet *Relation_sparse;
  BitString *Relation_dense;
  TIntSet *Deltas_sparse; // "Sparse Set"
  BitString *Deltas_dense; // "Dense Set"
 public:
  inline CGleichungszeile() /* Konstruktor */
    {
      relevanter_Faktor=no_factor;
      Relation_sparse = NULL;
      Relation_dense = NULL;
      Deltas_sparse = NULL;
      Deltas_dense = NULL;
    }

  inline void clear() /* Daten der Zeile lschen */
    {
      if (Relation_sparse) { delete Relation_sparse; Relation_sparse=NULL; }
      if (Relation_dense) { delete Relation_dense; Relation_dense=NULL; }
      if (Deltas_sparse) { delete Deltas_sparse; Deltas_sparse=NULL; }
      if (Deltas_dense) { delete Deltas_dense; Deltas_dense=NULL; }
      relevanter_Faktor=no_factor;
    }
  
  inline ~CGleichungszeile() /* Destruktor */
    { /* Besser nichts wirklich lschen, da dann das Kopieren
	 und Aufnehmen der Werte efizienter laeuft!
         Andernfalls: einen Zuweisungsoperator definieren, der
                      nicht die Zeiger kopiert, sondern deren
                      Inhalte! 
	 ---> Bisher also REFERENZSEMATIK implementiert! <---
      */
      //clear();
    }
  
  CGleichungszeile(unsigned int Delta_displacement); /* Konstruktur */
  /* "Deltas" mit dem Wert Delta initialisieren,
     "Relationen" mit den Primfaktoren des reduzierten Quadrates fllen,
     deren Exponenten ungerade sind. */

  inline void Wertkopie(const CGleichungszeile &GL)
    { /* wertsemantische Kopie von GL erzeugen! */
      if (GL.Relation_dense || GL.Deltas_dense)
	{
	  cerr << "Dense-Kopie noch nicht implementiert!" << endl;
	  exit(1);
	}
      Relation_dense=NULL; Deltas_dense=NULL;
      relevanter_Faktor=GL.relevanter_Faktor;
      Relation_sparse = new TIntSet; *Relation_sparse=*GL.Relation_sparse;
      Deltas_sparse = new TIntSet; *Deltas_sparse=*GL.Deltas_sparse;
    }

  int groesster_Faktor_in_Relation()
    /* (Index) des grter Faktor in der Relation */
    const
    {
      if (Relation_sparse)
	if (Relation_sparse->empty()) return no_factor;
        else 
	  {
	    TIntSet::iterator h = Relation_sparse->end();
	    return *--h;
	  }
      else return Relation_dense->last(1);
    }

  void Ausgabe(); /* Gleichungszeile ausgeben */
  void convert_Relation_to_dense(); /* mache "dense" aus "sparse" */
  void convert_Relation_to_sparse(); /* mache "sparse" aus "dense" */
  bool Relation_ist_leer() /* gibt zurck, ob die Relation leer ist */
    const
    {
      return (relevanter_Faktor==no_factor);
      // if (Relation_sparse) return Relation_sparse->empty();
      // else return Relation_dense->last(1)==-1; 
    }
  void convert_Deltas_to_dense(); /* mache "dense" aus "sparse" */
  void convert_Deltas_to_sparse(); /* mache "sparse" aus "dense" */	
  void kombinieren(const CGleichungszeile& GL2);
  /* die Vereinigungsmengen mit GL2 bilden und daraus die
     Schnittmengen entfernen. (-> also: symmetrische Differenz bilden)
     gibt zurck, ob die Relationsmenge leer ist, d.h. eine Kombination
     quadratischer Reste mit dem gleichen Wert gefunden wurde. */

  inline void swap(CGleichungszeile &GL2)
    { /* Inhalte der beiden Gleichungszeilen (referenzsemantisch) tauschen */
      TIntSet *h1;
      BitString *h2;
      h1=Relation_sparse; Relation_sparse=GL2.Relation_sparse;
      GL2.Relation_sparse=h1;
      h2=Relation_dense; Relation_dense=GL2.Relation_dense;
      GL2.Relation_dense=h2;
      h1=Deltas_sparse; Deltas_sparse=GL2.Deltas_sparse;
      GL2.Deltas_sparse=h1;
      h2=Deltas_dense; Deltas_dense=GL2.Deltas_dense;
      GL2.Deltas_dense=h2;
      int h3;
      h3=relevanter_Faktor; relevanter_Faktor=GL2.relevanter_Faktor;
      GL2.relevanter_Faktor=h3;
    }

  bool operator< (const CGleichungszeile &GL2)
    const
    {	
      return relevanter_Faktor < GL2.relevanter_Faktor;
    }
};

 
typedef list<CGleichungszeile> TZeilenliste;
typedef set <CGleichungszeile, less<CGleichungszeile> > TZeilenmenge;

#endif
